/*
 * Multiplication of complex arrays in interleaved form.
 * The interleaved form requires the real and imaginary parts of the complex number
 * to be stored in two consecutive locations of memory.
 */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>

#include <tmmintrin.h>

#include "utils.h"
#include "rdtsc.h"

#define N 1024

/* 
 * Procedure cplxmul: Serial code (do not need to modify)
 *
 * (a+ib)*(c+id) = (ac - bd) + i(ad + bc)
 *
 */
void cplxmul(float const * x, float const * y, float * z) {
  int i;
  for(i = 0; i < 2*N; i+=2) {
    z[i]   = x[i]*y[i] - x[i+1]*y[i+1];
    z[i+1] = x[i]*y[i+1] + x[i+1]*y[i];
  }

}



// Procedure vec_cplxmul: vector code (implement)

void vec_cplxmul(float const * x, float const * y, float * z) {

  // Substitute from here..
  printf("Implement vec_cplxmul\n");
  abort();
  // ...to here

}




/*
 * Do not need to modify from here on
 */

#define RUNS     400
#define CYCLES_REQUIRED 1e7

void verify(float const * x, float const * y, float const * z)
{
  int i;
  double err;
  float * temp = (float *) _mm_malloc(sizeof(float)*2*N, 16);
  setzero(temp, 2*N, 1);

  for(i = 0; i < 2*N; i+=2) {
      temp[i]   = x[i]*y[i] - x[i+1]*y[i+1];
      temp[i+1] = x[i]*y[i+1] + x[i+1]*y[i];
      err = fabs(z[i] - temp[i]);
      if(err > 1E-5)
        {
          printf("Error at z[%d] real\n", i/2);
        }
      err = fabs(z[i+1] - temp[i+1]);
      if(err > 1E-5)
        {
          printf("Error at z[%d] imag\n", i/2);
        }
  }

  _mm_free(temp);

}

void test_vec_cplxmul(float const * x, float const * y, float * z)
{
  tsc_counter start, end;
  double cycles = 0.;
  size_t num_runs = RUNS;
  int i;

  //Cache warm-up
  // RDTSCP reads ts register guaranteeing that the execution of all the code
  // we wanted to measure is completed. This way we avoid including the
  // execution of a CPUID in between. The last CPUID guarantees no other
  // instruction can be scheduled before it (and so also before RDTSCP)

  CPUID(); RDTSC(start); CPUID(); RDTSC(end);
  CPUID(); RDTSC(start); CPUID(); RDTSC(end);
  CPUID(); RDTSC(start); CPUID(); RDTSC(end);

  while(1) {
      CPUID(); RDTSC(start);
      for (i = 0; i < num_runs; ++i) {
          vec_cplxmul(x, y, z);
      }
      CPUID(); RDTSC(end);

      cycles = (double)(COUNTER_DIFF(end, start));

      if(cycles >= CYCLES_REQUIRED) break;

      num_runs *= 2;

  }

  CPUID(); RDTSC(start);
  for (i = 0; i < num_runs; ++i) {
      vec_cplxmul(x, y, z);
  }
  CPUID(); RDTSC(end);

  cycles = (double)(COUNTER_DIFF(end, start))/num_runs;

  printf("Test vec_cplxmul  - Performance [flops/cycle]: %f\n", 6.*N/cycles);

#ifdef VERIFY
  verify(x, y, z);
#endif

}

int main()
{
  float * x = (float *) _mm_malloc(sizeof(float)*2*N, 16);
  float * y = (float *) _mm_malloc(sizeof(float)*2*N, 16);
  float * z = (float *) _mm_malloc(sizeof(float)*2*N, 16);

  setrandom(x, 2*N, 1);
  setrandom(y, 2*N, 1);

  test_vec_cplxmul(x, y, z);

  return 0;
}
